#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import time
import subprocess
import collections

# ###########################################################
# torrentcycle.py, D.A. 12.+13.10.2013
#
# Dieses Script steuert den Start- / Stop-Status von Torrents auf einem
# Transmission-Client abhängig von der Auslastung.
# Komplett heruntergeladene Torrents werden gestoppt.
# ###########################################################

# Werte aus  /etc/transmission-daemon/settings.json
# "download-queue-enabled" muß auf false stehen, damit
# soviele Torrents gestartet werden können, wie dieses
# Programm will....und nicht Transmission irgendwann
# dicht macht.
maxPeersOverall=240                   # "peer-limit-global"
maxPeersPerTorrent=35                 # "peer-limit-per-torrent"
downloadDir="/nfs@11w/transmission"   # "download-dir"

log_ready="/home/dede/log_trans.txt"
log_status="/home/dede/log_torrent_cycle.txt"

HOST="bt"
AUTH="--auth=dede:passwort"


# ###########################################################
# Liefert ein Dictionary mit allen Elementen aus dem
# Output von "transmission-remote -l".
#
#          1         2         3         4         5         6         7
#01234567890123456789012345678901234567890123456789012345678901234567890123456789
#    |      |          |          |      |       |       |            |
#ID     Done       Have  ETA           Up    Down  Ratio  Status       Name
#   1    86%   891.7 MB  Unknown      0.0     0.0   5.10  Stopped      aaaaaaaaaa
#   3    99%    1.09 GB  Unknown     25.0     0.0   0.70  Uploading    bbbbbbbbbbb
#   5    97%    4.05 GB  13 hrs      77.0     0.0   2.14  Up & Down    cccccccccc
#   8    99%    1.09 GB  Unknown     97.0     0.0   2.85  Uploading    dddddddddddd
#   9    99%    1.39 GB  Unknown      0.0     0.0   0.94  Uploading    eeeeeeeeeee
#  10    99%    1.02 GB  Unknown      0.0     0.0   1.02  Idle         fffffffffff
#  13     0%    9.98 MB  Unknown      0.0     2.0   2.17  Up & Down    gggggggggggggg
#  16    46%   477.4 MB  Unknown      0.0     0.0   3.33  Uploading    hhhhhhhhhh
#  17     0%       None  Unknown      0.0     0.0   None  Uploading    iiiiiiiii
#Sum:          10.02 GB             199.0     2.0
def getDownloadList():
  cmd_l=["transmission-remote", HOST, AUTH, "-l"]
  process=subprocess.Popen(cmd_l, shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
  rc=process.communicate()[0]

  retdict={}
  if process.returncode==0:                 # nur bei rc==0 kann mit sinnvollen Daten gerechnet werden
    zeilenweise=rc.split("\n")              # Ausgabe in einzelne Zeilen aufteilen
    for idx in range(1, len(zeilenweise)-1):# über jede Zeile (ohne die Überschrift)
      ln=zeilenweise[idx].rstrip()
      if len(ln)>70:                        # nicht-Daten-Zeilen ausblenden
        t_id=int(ln[0:4].strip())
        t_name=ln[70:].strip()
        pd, av, pc, ts=getInfo(t_id)
        a={ "id"          : t_id,                 # int
            "done"        : pd,                   # float
            "have"        : ln[12:22].strip(),    # string  (987.4 MB)
            "eta"         : ln[23:33].strip(),    # string  (Unknown) (2 hrs)
            "up"          : ln[34:40].strip(),    # string  (22.0)
            "down"        : ln[41:48].strip(),    # string  (13.0)
            "ratio"       : ln[49:56].strip(),    # string  (None) (1.30)
            "status"      : ln[57:69].strip(),    # string  (Stopped) (Up & Down) (Uploading) (Downloading) (Finished) (Idle)
            "name"        : t_name,               # string  (<irgendwas mit oder ohne Leerzeichen>)
            "connected"   : pc,                   # int
            "avail"       : av,                   # float
            "size_wanted" : ts                    # string  (1.80 GB)
            }
        retdict.update({t_name : a})
  return(retdict)


# ###########################################################
# Liefert von dem Torrent mit der ID "t_id" die Daten
#   "Percent Done", 
#   "Availability",
#   "Peers connected",
#   "Total size (wanted)"
# als Tupel.
def getInfo(t_id):
  cmd_l=["transmission-remote", HOST, AUTH, "-t", str(t_id), "-i"]
  process=subprocess.Popen(cmd_l, shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
  rc=process.communicate()[0]

  pd=av=pc=0  # falls der Torrent just in dem Moment zwischen "-l" und "-t x -i" gelöscht wird, ...
  ts="0 GB"   # ...werden hier default-Werte gesetzt.
  if process.returncode==0:
    zeilenweise=rc.split("\n")
    for ln in zeilenweise:
      lns=ln.strip()
      if lns.find("Percent Done:")==0:
        # Percent Done: 46.1%
        #               ----
        try:
          pd=float(lns[13:].strip("% "))
        except ValueError:
          pass
      if lns.find("Availability:")==0:
        # Availability: 100%
        #               ---
        try:
          av=float(lns[13:].strip("% "))
        except ValueError:
          pass
      if lns.find("Peers: connected to")==0:
        # Peers: connected to 33, uploading to 13, downloading from 7
        #                     --
        try:
          pc=int(lns[19:].strip().split()[0].rstrip(","))
        except ValueError:
          pass
      if lns.find("Total size:")==0:
        # Total size: 1.07 GB (1.06 GB wanted)
        #                      ---- --
        try:
          tst=lns[11:].strip().split()
          ts=tst[2].lstrip("(") + " " + tst[3]
        except IndexError:
          pass
  return((pd, av, pc, ts))


# ###########################################################
# Stoppt den Torrent mit der ID "t_id".
def stopTorrent(t_id):
  cmd_d=["transmission-remote", HOST, AUTH, "-t"+str(t_id), "-S"]
  process=subprocess.Popen(cmd_d, shell=False, stdout=subprocess.PIPE)
  rc=process.communicate()[0]

# ###########################################################
# Startet den Torrent mit der ID "t_id".
#
# braucht ein
#   "download-queue-enabled": false,
# in /etc/transmission-daemon/settings.json
def startTorrent(t_id):
  cmd_d=["transmission-remote", HOST, AUTH, "-t"+str(t_id), "-s"]
  process=subprocess.Popen(cmd_d, shell=False, stdout=subprocess.PIPE)
  rc=process.communicate()[0]


# ###########################################################
# Schreibt "t_name" samt Datum+Uhrzeit in die Datei "log_ready".
def logTorrentCompletion(t_name):
  fobj=open(log_ready, "a")
  fobj.write(time.strftime("%d.%m.%Y %H:%M:%S")+" "+t_name+"\n")
  fobj.close()

# ###########################################################
# Schreibt "txt" in die Datei "log_status".
def logStatus(txt):
  fobj=open(log_status, "a")
  fobj.write(txt+"\n")
  fobj.close()


# ###########################################################
# Liefert False, wenn der Torrent mit der Größe "sizeTorrent"
# nicht mehr gestartet werden kann.
#
# sizeTorrent im Format "1.35 GB" oder "739.7 MB"
# kann auch "None wanted)" sein -> dann wird False geliefert
def checkSize(sizeTorrent):
  sz=os.statvfs(downloadDir)      # Daten der Platte holen
  free=sz.f_bsize*sz.f_bfree      # Restkapazität der Platte
  st=sizeTorrent.split()          # Torrent-Größe nach Wert und Größenordnung wandeln
  logStatus("sizeTorrent="+sizeTorrent)
  if len(st)!=2:                  # nicht das gewünschte Format...
    return(False)                 # ...dann wirds auch nicht gestartet
  if st[1]=="kB":                 # Größenordnung nach Zahl
    m=1024.0
  elif st[1]=="MB":
    m=1024.0*1024.0
  elif st[1]=="GB":
    m=1024.0*1024.0*1024.0
  else:                           # keine Größenordnung erkannt...
    return(False)                 # ...dann wirds auch nicht gestartet
  keepFree=1024.0*1024.0*1024.0*4.0     # vier GigaByte frei lassen
  if ((free-keepFree)/m)<float(st[0]):  # wenn Restkapazität kleiner als Torrent-Größe...
    return(False)                       # "Passt nicht" liefern
  return(True)


# ###########################################################
# Schreibt den Inhalt der als "que" übergebenen [de]Queue
# hübsch formatiert ins Log. Der in "prefix" übergebene
# Text wird links vor dem dem Queue-Inhalt ausgeben.
def printQueueToLog(que, prefix):
  od=""                       # zum Aufbau einer Zeile
  odl=[]                      # alle Zeilen
  for i in que:               # über alle Einträge
    if (len(od)+len(i))>70:   # wenn Zeilenlänge zu lang werden würde...
      odl.append(od)          # bisherige Zeile als fertig ablegen
      od=i+" "                # der aktuelle Eintrag landet in der nächsten Zeile
    else:                     # wenn Zeilenlänge passt...
      od+=i+" "               # aktuellen Eintrag anhängen
  odl.append(od)              # letzte Zeile als fertig ablegen
  for i in range(len(odl)):             # über alle Zeilen
    if i==0:                            # bei der ersten Zeile...
      logStatus(prefix+odl[i])          # ... vorab den prefix ausgeben
    else:                               # nach der ersten Zeile...
      logStatus(" "*len(prefix)+odl[i]) # ...nur noch analog zum prefix ausrichten




# ###########################################################
# #################                         #################
# ################       Hauptprogramm       ################
# #################                         #################
# ###########################################################

startQueue=collections.deque()    # hier werden die Namen der gestoppten Torrents in einem FIFO abgelegt
stoppedQueue=collections.deque()  # hier werden die Namen der fertigen Torrents in einem FIFO abgelegt

rl=getDownloadList()
hist={}   # Key ist der Torrent-Name, Value ist eine Liste aus "have", "done" und dem Stop-Zähler
for k, v in rl.items():
  hist.update({k : [v["have"], v["done"], 0]})    # initiales Füttern der Historie
  if v["status"]=="Stopped":                      # gestoppte, ....
    if v["done"]!=100:                            # ...nicht fertige Torrents werden...
      startQueue.append(k)                        # ...auf der start-Queue abgelegt
    else:
      stoppedQueue.append(k)                      # ...fertige hingegen auf der stopped-Queue

lst_format="{0:25} {1:2} {2:>10} {3:5}% {4:5}% {5:>11} {6:2} {7:>8}"
# 0: Torrent-Name
# 1: stop-Zähler
# 2: bereits empfangene Dateigröße
# 3: bereits empfangene Prozent
# 4: verfügbare Prozent
# 5: Lade-Status
# 6: verbundene Peers
# 7: herunterzuladende Dateigröße

while True:
  time.sleep(60)          # Status im Minutentakt prüfen
  rl=getDownloadList()    # neuen Status von Transmission holen
  actions=[]              # Anzeigetexte

  # ------------------------------------------------------------------------
  # verbundene Peer zählen
  # ------------------------------------------------------------------------
  currPeersOverall=0
  for k, v in rl.items():
    currPeersOverall+=v["connected"]  # zählt die Anzahl der verbundenen Peers

  # ------------------------------------------------------------------------
  # Restkapazität der Platte holen
  # ------------------------------------------------------------------------
  sz=os.statvfs(downloadDir)                    # Daten der Platte holen
  hddfree=sz.f_bsize*sz.f_bfree/1073741824.0    # Restkapazität der Platte in GB

  # ------------------------------------------------------------------------
  # Überschrift anzeigen
  # ------------------------------------------------------------------------
  logStatus("")                                   
  logStatus(time.strftime("%d.%m.%Y %H:%M:%S")+" "+"-"*16+" HDD: {0:3.1f} GB ".format(hddfree)+ \
          "-"*16+" Peers: " + str(currPeersOverall))

  # ------------------------------------------------------------------------
  # die Queues bereinigen, falls Torrents aus Transmission gelöscht wurden
  # ------------------------------------------------------------------------
  # die aus transmission gelöschten Torrents aus der stoppedQueue löschen
  lst=[]
  for l in stoppedQueue:
    if l not in rl:
      lst.append(l)         # erst zwischenspeichern (um die for-Schleife nicht zu verwirren)
  for l in lst:
    stoppedQueue.remove(l)  # und dann alle gemerkten Torrents löschen

  # die aus transmission gelöschten Torrents aus der startQueue löschen
  lst=[]
  for l in startQueue:
    if l not in rl:
      lst.append(l)       # erst zwischenspeichern (um die for-Schleife nicht zu verwirren)
  for l in lst:
    startQueue.remove(l)  # und dann alle gemerkten Torrents löschen

  # ------------------------------------------------------------------------
  # über alle Torrents (sortiert nach Name)
  # ------------------------------------------------------------------------
  for k, v in sorted(rl.items(), key=lambda x: x[0]):
    # ------------------------------------------------------------------------
    # neu zugefügte oder manuell gestoppte Torrents ggf. auf startQueue ablegen
    # ------------------------------------------------------------------------
    if v["status"]=="Stopped":    # wenn Torrent gestoppt...
      if v["done"]!=100:          # und nicht komplett heruntergeladen ist...
        if k not in startQueue:   # und noch nicht auf der startQueue...
          startQueue.append(k)    # dann rauf damit

    # ------------------------------------------------------------------------
    # laufende Torrents von der startQueue nehmen
    # ------------------------------------------------------------------------
    if v["status"]!="Stopped":    # wenn Torrent nicht gestoppt...
      if k in startQueue:         # und auf der startQueue...
        startQueue.remove(k)      # dann weg damit

    # ------------------------------------------------------------------------
    # komplette/fertige Torrents ggf. stoppen und auf stoppedQueue ablegen
    # ------------------------------------------------------------------------
    if v["done"]==100:                    # wenn Torrent komplett heruntergeladen ist...
      if k not in stoppedQueue:           # und nicht eh schon als fertig markiert ist
        stopTorrent(v["id"])              # dann stoppen
        currPeersOverall-=v["connected"]  # Anzahl verbundener Peers anpassen
        stoppedQueue.append(k)            # und als fertig markieren
        actions.append("komplett ist: " + k)
        logTorrentCompletion(k)

    if k not in hist:                               # wenn Torrent frisch aufgenommen wurde
      hist.update({k : [v["have"], v["done"], 0]})  # erstmal history anlegen
    else:
      # ------------------------------------------------------------------------
      # stop-zähler korrigieren bei gestopptem Torrent
      # ------------------------------------------------------------------------
      if v["status"]=="Stopped":      # frisch gestoppter Torrent
        if hist[k][2]>0:              # soll immer einen stop-zähler == 0 haben
          hist[k][2]=0

      # ------------------------------------------------------------------------
      # stop-zähler entsprechend Download-Verhalten des Torrents setzen
      # ------------------------------------------------------------------------
      if hist[k][0]==v["have"] and hist[k][1]==v["done"]: # wenn seit letztem Durchlauf nix heruntergeladen wurde...
        if v["status"]!="Stopped":                        # und nicht gestoppt
          if k not in stoppedQueue:                       # oder auf sowas wie "Finished" steht und trotzdem als fertig markiert ist
            hist[k][2]+=1                                 # stop-zähler inkrementieren
      else:                                               # wenn heruntergeladen wurde...
        hist.update({k : [v["have"], v["done"], 0]})      # stop-zähler wieder auf 0 und neue Werte merken

    # ------------------------------------------------------------------------
    # Anzeige der aktiven Torrents
    # ------------------------------------------------------------------------
    if v["status"]!="Stopped":  # erstmal nur die aktiven Torrents anzeigen
      logStatus(lst_format.format(k, hist[k][2], v["have"], v["done"], v["avail"], v["status"], v["connected"], v["size_wanted"]))

    # ------------------------------------------------------------------------
    # schon seit einigen Durchläufen nicht mehr ladende Torrents stoppen
    # ------------------------------------------------------------------------
    if len(startQueue)>4: # je nach Queue-Länge die Anzahl der Schleifen-Durchläufe ohne Download festlegen
      cyc=4     # bei langer Queue: 4 leere Durchläufe abwarten
    else:
      cyc=8     # bei kurzer Queue: 8 leere Durchläufe abwarten
    if hist[k][2]>cyc:                          # wenn zu lange nix heruntergeladen wurde
      if len(startQueue)>0:                     # und es Warte-Kandidaten gibt
        if k not in stoppedQueue:               # und nicht eh schon als fertig markiert
          stopTorrent(v["id"])                  # dann stoppen
          currPeersOverall-=v["connected"]      # Anzahl verbundener Peers anpassen
          actions.append("gestoppt wurde (download timeout): " + k)

    # ------------------------------------------------------------------------
    # Torrents wegen mangelnder Verfügbarkeit stoppen
    # ------------------------------------------------------------------------
    if v["status"]!="Stopped":                            # wenn Torrent nicht gestoppt
      if k not in stoppedQueue:                           # und Torrent nicht als fertig markiert
        if len(startQueue)>0:                             # und es Warte-Kandidaten gibt
          if currPeersOverall>(maxPeersOverall*0.9):      # und reichlich Peers verbunden sind
            if v["connected"]>=(maxPeersPerTorrent*0.9):  # und dieser Torrent fast das Maximum an Peers bindet
              if (v["done"]+0.1)>v["avail"]:              # und fast alle verfügbaren Teile da sind
                if v["avail"]<100:                        # aber nicht 100% verfügbar sind
                  stopTorrent(v["id"])                    # dann stoppen
                  currPeersOverall-=v["connected"]        # Anzahl verbundener Peers anpassen
                  actions.append("gestoppt wurde (Torrent-Bindung): " + k)

  # ------------------------------------------------------------------------
  # starten eines Torrents pro Hauptschleifen-Durchlauf
  # ------------------------------------------------------------------------
  if currPeersOverall<(maxPeersOverall*0.9):          # wenn noch Peer-Slots frei sind
    if len(startQueue)>0:                             # und es Warte-Kandidaten gibt
      t_nam=startQueue.popleft()                      # den ältesten Torrent von der Queue holen
      if rl[t_nam]["have"]=="None":                   # wenn Torrent bisher keinen Platz belegt...
        if checkSize(rl[t_nam]["size_wanted"])==True: # ...Platz prüfen
          startTorrent(rl[t_nam]["id"])               # Platz ist da -> starten
          actions.append("gestartet wurde: " + t_nam)
        else:
          # der von der startQueue genommene Torrent wird oben (beim nächsten Durchlauf) wieder der startQueue angehängt
          actions.append("nicht gestartet wurde (speicherplatz): " + t_nam)
      else:
        startTorrent(rl[t_nam]["id"])               # Torrent belegt eh schon Platz auf der Platte -> also starten
        actions.append("gestartet wurde: " + t_nam)


  # ------------------------------------------------------------------------
  # Anzeige der gestoppten Torrents, Aktionen und Queues
  # ------------------------------------------------------------------------
  for k, v in sorted(rl.items(), key=lambda x: x[0]): # die gestoppten Torrents (nach Name sortiert) anzeigen
    if v["status"]=="Stopped":
      logStatus(lst_format.format(k, hist[k][2], v["have"], v["done"], v["avail"], v["status"], v["connected"], v["size_wanted"]))

  # Anzeige der Aktionen
  for i in actions:
    logStatus("   --   "+i)

  # Anzeige der Queues
  printQueueToLog(startQueue,   "start : ")   # die Liste der Warte-Kandidaten
  printQueueToLog(stoppedQueue, "fertig: ")   # und die fertigen Torrents

